package com.vernaio.example;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.supercsv.io.CsvListReader;
import org.supercsv.prefs.CsvPreference;

import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.fasterxml.jackson.datatype.jdk8.Jdk8Module;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoBooleanDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoFloatDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoBooleanDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoSegmentFloatValue;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoSegmentsFloatDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.exception.EngineException;
import com.vernaio.aivis.engine.v2.base.sdk.exception.MissingLibraryException;
import com.vernaio.aivis.engine.v2.base.sdk.exception.WrongFlavourException;
import com.vernaio.aivis.engine.v2.state.detection.sdk.StateDetectionData;

public class Utils {
	
	private static final ObjectMapper MAPPER = new ObjectMapper().registerModule(new Jdk8Module());
	
	public static abstract class Signal {
		
		private final String signalId;
		
		protected Signal(final String signalId) {
			this.signalId = signalId;
		}
		
		public String getSignalId() {
			return this.signalId;
		}
	}
	
	public static class FloatSignal extends Signal {
		
		private final List<DtoFloatDataPoint> dataPoints;
		
		protected FloatSignal(final String signalId) {
			super(signalId);
			this.dataPoints = new ArrayList<DtoFloatDataPoint>();
		}
		
		public List<DtoFloatDataPoint> getDataPoints() {
			return this.dataPoints;
		}
	}
	
	public static class BooleanSignal extends Signal {
		
		private final List<DtoBooleanDataPoint> dataPoints;
		
		protected BooleanSignal(final String signalId) {
			super(signalId);
			this.dataPoints = new ArrayList<DtoBooleanDataPoint>();
		}
		
		public List<DtoBooleanDataPoint> getDataPoints() {
			return this.dataPoints;
		}
	}
	
	/*
	 * The following method knows very well how the data csv in this example looks like, reads it and at the end constructs the data context:
	 * Signals are added to the data context by naming their signal id and a list of their data points (DtoFloatDataPoints for float signals, analogously for
	 * boolean and string signals).
	 * The method can be adapted if the data comes in a different format, e.g. via parquet files.
	 */
	public static StateDetectionData readCsvConvertToAivisFormat(final File file, final String targetSignalId) throws FileNotFoundException, IOException,
		WrongFlavourException, MissingLibraryException, EngineException {
		final HashMap<String, Signal> signals = new HashMap<String, Signal>();
		final CsvPreference csvPreference = new CsvPreference.Builder('"', ',', "\n").ignoreEmptyLines(false).build();
		
		System.out.println(String.format("\t... reading file '%s' ...", file.getPath()));
		
		// get line numbers to show nice progress bar
		int lineCount = 0;
		try (LineNumberReader count = new LineNumberReader(new FileReader(file))) {
			count.skip(Long.MAX_VALUE);
			lineCount = count.getLineNumber();
		}
		
		// read file and construct lists of data points
		try (
			final InputStream stream = new FileInputStream(file);
			final InputStreamReader reader = new InputStreamReader(stream, StandardCharsets.UTF_8);
			final CsvListReader csvReader = new CsvListReader(reader, csvPreference)
		) {
			
			// the signals ids are in the csv's header, the first column are the timestamps
			final String[] header = csvReader.getHeader(true);
			for (int i = 1; i < header.length; i++) {
				final String signalId = header[i];
				if (signalId.contentEquals(targetSignalId)) {
					signals.put(signalId, new BooleanSignal(signalId)); // the target is a boolean signal...
				} else {
					signals.put(signalId, new FloatSignal(signalId)); // ... while all other signals have float values
				}
			}
			
			System.out.print("\t... processing signals ...");
			
			// show nice progress bar
			int progressBar = -1;
			int line = 0;
			
			// read values by index
			List<String> values;
			while ((values = csvReader.read()) != null) {
				
				// show nice progress bar
				final int progressBarNew = line / (lineCount / 10) * 10;
				if (progressBar != progressBarNew) {
					progressBar = progressBarNew;
					// System.out.print("\r");
					System.out.print(String.format("\r\t... processing signals: %d%% ...", progressBar));
				}
				
				// timestamp is in the first column
				final long timestamp = Long.parseLong(values.get(0));
				
				// to this timestamp, add dataPoints to all signals (if value exists)
				for (int i = 1; i < values.size(); i++) {
					final String stringValue = values.get(i);
					if (stringValue != null && !stringValue.isEmpty()) {
						final double doubleValue = Double.parseDouble(stringValue);
						final String signalId = header[i];
						if (signalId.contentEquals(targetSignalId)) {
							((BooleanSignal) signals.get(signalId)).getDataPoints().add(new DtoBooleanDataPoint(timestamp, doubleValue > 0.5));
						} else {
							((FloatSignal) signals.get(signalId)).getDataPoints().add(new DtoFloatDataPoint(timestamp, doubleValue));
						}
					}
				}
				
				line++;
			}
		}
		
		System.out.print(String.format("\r\t... processing signals: %d%% ...", 100));
		
		// finally create data context
		System.out.println("and now adding signals to data context:");
		final StateDetectionData stateDetectionData = StateDetectionData.create();
		
		for (Map.Entry<String, Signal> signal : signals.entrySet()) {
			if (signal.getValue() instanceof FloatSignal) {
				stateDetectionData.addFloatSignal(signal.getKey(), ((FloatSignal) signal.getValue()).getDataPoints());
				System.out.println(String.format("\t... float signal '%s' has %s values", signal.getKey(), ((FloatSignal) signal.getValue()).getDataPoints().size()));
			} else if (signal.getValue() instanceof BooleanSignal) {
				stateDetectionData.addBooleanSignal(signal.getKey(), ((BooleanSignal) signal.getValue()).getDataPoints());
				System.out.println(
					String.format("\t... boolean signal '%s' has %s values", signal.getKey(), ((BooleanSignal) signal.getValue()).getDataPoints().size())
				);
			} else {
				throw new IOException("Unknown DataType!");
			}
		}
		System.out.println("\t... data context is finished!");
		
		return stateDetectionData;
	}
	
	public static <T> T readJson(final File file, final Class<T> clazz) throws JsonParseException, JsonMappingException, IOException {
		System.out.println(String.format("\t... reading file '%s' ...", file.getPath()));
		return MAPPER.readValue(file, clazz);
	}
	
	public static void writeAdditionalSignalsTimeseriesWithTarget(
		final File file,
		final List<? extends IDtoBooleanDataPoint> targetDataPoints,
		final String targetSignalId
	) throws FileNotFoundException, IOException {
		System.out.println(String.format("\t... writing file '%s' ...", file.getPath()));
		
		final ArrayNode json = MAPPER.createArrayNode();
		final ObjectNode timeseriesByIdNode = MAPPER.createObjectNode();
		json.add(timeseriesByIdNode);
		
		timeseriesByIdNode.put("id", targetSignalId);
		
		final ObjectNode timeseriesNode = MAPPER.createObjectNode();
		timeseriesByIdNode.set("timeseries", timeseriesNode);
		
		timeseriesNode.put("_type", "Boolean");
		final ArrayNode dataPointsNode = timeseriesNode.putArray("dataPoints");
		for (IDtoBooleanDataPoint dataPoint : targetDataPoints) {
			final ObjectNode dataPointNode = MAPPER.createObjectNode();
			dataPointNode.put("timestamp", dataPoint.getTimestamp());
			dataPointNode.put("value", dataPoint.getValue());
			dataPointsNode.add(dataPointNode);
		}
		
		final FileWriter fileWriter = new FileWriter(file);
		fileWriter.write(json.toPrettyString());
		fileWriter.close();
		
		return;
		
	}
	
	public static void writeSegmentTimeseries(final File file, final List<? extends IDtoSegmentsFloatDataPoint> timeseries) throws FileNotFoundException,
		IOException {
		System.out.println(String.format("\t... writing file '%s' ...", file.getPath()));
		
		final ObjectNode json = MAPPER.createObjectNode();
		
		json.put("_type", "SegmentsFloat");
		final ArrayNode dataPointsNode = json.putArray("dataPoints");
		for (IDtoSegmentsFloatDataPoint dataPoint : timeseries) {
			final ObjectNode dataPointNode = MAPPER.createObjectNode();
			dataPointNode.put("timestamp", dataPoint.getTimestamp());
			final ArrayNode valuesNode = dataPointNode.putArray("values");
			for (IDtoSegmentFloatValue segmentValue : dataPoint.getValues()) {
				final ObjectNode valueNode = MAPPER.createObjectNode();
				valueNode.put("segmentId", segmentValue.getSegmentId());
				valueNode.put("value", segmentValue.getValue());
				valuesNode.add(valueNode);
			}
			
			dataPointsNode.add(dataPointNode);
		}
		
		final FileWriter fileWriter = new FileWriter(file);
		fileWriter.write(json.toPrettyString());
		fileWriter.close();
		
	}
	
	public static <T> void writeJson(final File file, final T obj) throws JsonGenerationException, JsonMappingException, IOException {
		System.out.println(String.format("\t... writing file '%s' ...", file.getPath()));
		Optional.ofNullable(file.getParentFile()).ifPresent(f -> f.mkdirs());
		MAPPER.writeValue(file, obj);
	}
	
}
